from math import pi
from casadi import *
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
from scipy.stats import norm 
from scipy.stats import invgauss
from collections import OrderedDict 
from helpers import casadi_vec, casadi_struct, casadi_vec2struct, casadi_struct2vec 
from Parameters import Parameters # Script for all of the parameters
from Model import Model # Script for the model equations
from Layout import Layout # Script for structure of the DAE system
from demopt import demopt # Deterministic optimization script
from Initial import Initial # The initial guess script

p = Parameters();
[f,dx,obj,x,z,u,t_end] = Model(p);
[Int,dae,dae,dae_x,dae_z,dae_u,tau_root,dae_t_end] = Layout(p,f,dx,obj,x,z,u,t_end);
[v_opt,nvar,lbx,ubx,tf] = demopt(dae_x,dae_z,dae_u,dae,p,z);

p.TF = float(v_opt[-1])*p.TF
p.On = 0

[f,dx,obj,x,z,u,t_end] = Model(p);
[Int,dae,dae,dae_x,dae_z,dae_u,tau_root,dae_t_end] = Layout(p,f,dx,obj,x,z,u,t_end);
# Import collocation script
from collocation_tools import simpleColl
if p.On == 1:
	from collocation_tools_t import simpleColl
else:
	from collocation_tools import simpleColl
collfun = simpleColl(dae,tau_root,1.0/p.N)
collfun = collfun.expand()


def MonteCarlo(p,v_opt,nvar):
	"""This is a function that runs the Monte Carlo simulations.
	"""
	U_init = []
	for i in range(p.N):
		U_init.append(v_opt[(p.tz+i*nvar):(p.tz+3+i*nvar)])

	U_init = vertcat(*U_init)
	X_out1 = []
	Z_out1 = []
	for i in range(p.ns):
	
		# Generates random outcomes for ko
		if i == 0:
			ko = p.ko
		mu_ko = ko # Mean value of the reservoir flow coefficient of oil
		sigma_ko = 0.1*ko # Variance of the reservoir flow coefficient of oil
		k_o = np.random.normal(mu_ko, sigma_ko)
		p.ko = k_o # 
		
		# Update the model with new value of ko
		
		[f,dx,obj,x,z,u,t_end] = Model(p);

		dae_int = {'x':x, 'z':z, 'p':u,  'ode':dx, 'alg':f, 'quad':obj}

		opts = {}
		opts['abstol'] = 1e-8# abs. tolerance
		opts['reltol'] = 1e-8 # rel. tolerance
		opts['linear_solver'] = 'csparse'
		opts['linear_solver_type'] = 'user_defined'
		opts['tf'] = 1.0/p.N
		INT = integrator('INT', 'idas', dae_int, opts)
		
		X_out = []
		Z_out = []
		for j in range(p.N):

			if j == 0:

				sim = INT( x0=v_opt[0:2], z0=v_opt[p.tx:p.tx+34], p=U_init[0:3])
				xfin = sim['xf']
				zfin = sim['zf']

				X_init = xfin
				Z_init = zfin
				X_out.append(X_init)
				Z_out.append(Z_init)
			else:

				sim = INT( x0=X_init, z0=Z_init, p=U_init[j*3:j*3+3])
				xfin = sim['xf']
				zfin = sim['zf']
				X_init = xfin
				Z_init = zfin

				
				X_out.append(X_init)
				Z_out.append(Z_init)
		
		X_out1.append(vertcat(*X_out))
		Z_out1.append(vertcat(*Z_out))

	X_sim = np.hstack(X_out1)
	Z_sim = np.hstack(Z_out1)
	p.ko = ko
	
	return X_sim,Z_sim				


def backoff(X_out,Z_out):
	""" This script calculates the backoff for the chance constraint
	"""
	alpha_eta = 0.95 # probability level of eta
	alpha_Z3_g = 0.95 # probability level of Z5_o
	alpha_P_6 = 0.95 # probability level of P_6

	mu_eta = np.zeros(p.N) # Make the mean value matrix
	mu_Z3_g = np.zeros(p.N) # Make the mean value matrix
	mu_P_6 = np.zeros(p.N) # Make the mean value matrix
	
	sigma_eta = np.zeros(p.N) # Make the variance matrix
	sigma_Z3_g = np.zeros(p.N) # Make the variance matrix
	sigma_P_6 = np.zeros(p.N) # Make the variance matrix

	back_off_eta = np.zeros(p.N) # Make the back_off matrix for eta
	back_off_Z3_g = np.zeros(p.N) # Make the back_off matrix for Z5_o
	back_off_P_6 = np.zeros(p.N) # Make the back_off matrix for P_6
	
	for i in range(p.N):
		
		# Calculates the back off for each timestep
		[mu_eta[i], sigma_eta[i]] = norm.fit(X_out[i*2+1,:]) # Normalfit the outcome of eta
		[mu_Z3_g[i], sigma_Z3_g[i]] = norm.fit(Z_out[i*34+15,:]) # Normalfit the outcome of Z3_g
		[mu_P_6[i], sigma_P_6[i]] = norm.fit(Z_out[i*34+33,:]) # Normalfit the outcome of P_6
		

		back_off_eta[i] = norm.ppf(alpha_eta)*sqrt(np.cov(X_out[i*2+1,:])) # calculates the back off of eta

		back_off_Z3_g[i] = norm.ppf(alpha_Z3_g)*sqrt(np.cov(Z_out[i*34+15,:])) # calculates the back off of Z3_g

		back_off_P_6[i] =  norm.ppf(alpha_P_6)*sqrt(np.cov(Z_out[i*34+33,:])) # calculates the back off of P_6

	return back_off_eta, back_off_Z3_g,back_off_P_6,mu_eta,mu_Z3_g,mu_P_6,sigma_eta,sigma_Z3_g,sigma_P_6
	
[X_out,Z_out] = MonteCarlo(p,v_opt,nvar);

[back_off_eta, back_off_Z3_g,back_off_P_6,mu_eta,mu_Z3_g,mu_P_6,sigma_eta,sigma_Z3_g,sigma_P_6] = backoff(X_out,Z_out);

BO_eta = back_off_eta.tolist()
BO_Z3_g = back_off_Z3_g.tolist()
BO_P6 = back_off_P_6.tolist()

def StochOpt(p,v_opt,nvar,BO_P6,BO_Z3_g,BO_eta):
	
	# Import collocation script
	from collocation_tools import simpleColl
	if p.On == 1:
		from collocation_tools_t import simpleColl
	else:
		from collocation_tools import simpleColl
	collfun = simpleColl(dae,tau_root,1.0/p.N)
	collfun = collfun.expand()
	# Number of varibles
	nx = dae["x"].shape[0] # Differential variables
	nu = dae["p"].shape[0] # Control variables
	nz = dae["z"].shape[0] # Algebraic variables
	nvar = nx + nx*p.d + nz*p.d + nu # number of variables 
	nv = nx*(p.N+1) + nx*p.d*p.N + nz*p.d*p.N + nu*p.N # Total number of variables
	
	Xs  = [MX.sym("X",nx) for i in range(p.N+1)]
	XCs = [MX.sym("XC",nx,p.d) for i in range(p.N)]
	Zs  = [MX.sym("Z",nz,p.d) for i in range(p.N)]
	Us  = [MX.sym("U",nu) for i in range(p.N)]
	if p.On == 1:
		t_end = MX.sym("t_end")
	V_block = OrderedDict()
	V_block["X"]  = Sparsity.dense(nx,1)
	V_block["XC"] = Sparsity.dense(nx,p.d)
	V_block["Z"]  = Sparsity.dense(nz,p.d)
	V_block["U"]  = Sparsity.dense(nu,1)
	
	# Simple bounds on states
	lbx = [];
	ubx = [];

	# Objective function
	f = 0.0;

	# List of constraints
	g = [];
	
	# List of all decision variables (determines ordering)
	V = [];
	
	for k in range(p.N):
		# Add decision variables
		V += [casadi_vec(V_block,X=Xs[k],XC=XCs[k],Z=Zs[k],U=Us[k])]
		
		if k==0:
			# Bounds at t=0
			x_lb = casadi_vec(dae_x,0.0,Pr=1.0,eta=1.0)
			x_ub = casadi_vec(dae_x,1.0,Pr=1.0,eta=1.0)
			xc_lb = casadi_vec(dae_x,0.0,eta=0.80)
			xc_lb = repmat(xc_lb,1,p.d)
			xc_ub = casadi_vec(dae_x,1.0)
			xc_ub = repmat(xc_ub,1,p.d)
			u_lb = casadi_vec(dae_u,0.0,P2=0.2,h=0.5)
			u_ub = casadi_vec(dae_u,1.0,h=2.5)
			z_lb = casadi_vec(dae_z,0.0,P6=0.48 )
			z_lb = repmat(z_lb,1,p.d)
			z_ub = casadi_vec(dae_z,1.0,Z5_o=0.01,Z3_g=0.1 - BO_Z3_g[k])
			z_ub = repmat(z_ub,1,p.d)
			lbx.append(casadi_vec(V_block,-inf,X=x_lb,XC=xc_lb,Z=z_lb,U=u_lb))
			ubx.append(casadi_vec(V_block,inf,X=x_ub,XC=xc_ub,Z=z_ub,U=u_ub))
		else:
			# Bounds for other t
			x_lb = casadi_vec(dae_x,0.0,eta=0.80)
			x_ub = casadi_vec(dae_x,1.0)
			xc_lb = casadi_vec(dae_x,0.0,eta=0.80)
			xc_lb = repmat(xc_lb,1,p.d)
			xc_ub = casadi_vec(dae_x,1.0)
			xc_ub = repmat(xc_ub,1,p.d)
			u_lb = casadi_vec(dae_u,0.0,P2=0.2,h=0.5)
			u_ub = casadi_vec(dae_u,1.0,h=2.5)
			z_lb = casadi_vec(dae_z,0.0,P6=0.48 )
			z_lb = repmat(z_lb,1,p.d)
			z_ub = casadi_vec(dae_z,1.0,Z5_o=0.01,Z3_g=0.1 - BO_Z3_g[k])
			z_ub = repmat(z_ub,1,p.d)
			lbx.append(casadi_vec(V_block,-inf,X=x_lb,XC=xc_lb,Z=z_lb,U=u_lb))
			ubx.append(casadi_vec(V_block,inf,X=x_ub,XC=xc_ub,Z=z_ub,U=u_ub))
		# Obtain collocation expressions
		if p.On == 1:
			# Varying time horizon
			[Xnext,constr,quad_k] = collfun(Xs[k],XCs[k],Zs[k],Us[k],t_end)
		else:
			# Fixed time horizon
			[Xnext,constr,quad_k] = collfun(Xs[k],XCs[k],Zs[k],Us[k])
		g.append(constr) # Add algebraic constraints
		g.append(Xs[k+1] - Xnext) # Add continuity constraint added
		# Add to the objective function
		if p.PV == 1:
			# Present Value is included
			f += quad_k/(1.05)**(k*1.0/12.0)
		else:
			# Present Value is not included
			f += quad_k
		
	V += [Xs[-1]] 
	
	if p.On == 1:
		# Add time horizon as varible for varying time horizin
		V += [t_end]
		f *= V[-1] 
		
	# Bounds for final t
	x_lb = casadi_vec(dae_x,0.0,eta=0.8 + BO_eta[-1])
	x_ub = casadi_vec(dae_x,1.0)
	lbx.append(x_lb)
	ubx.append(x_ub)
	
	if p.On == 1:
		# Add constraint on end time horizon
		t_lb = casadi_vec(dae_t_end,0.8)
		t_ub = casadi_vec(dae_t_end,1.2)
		lbx.append(t_lb)
		ubx.append(t_ub)
	
	#Initial guess
	init = Initial(p,z,V_block);
	if p.On == 1:
		V0 = np.zeros(init.shape[0]+1)
	else:
		V0 = np.zeros(init.shape[0])
		
	for q in range(init.shape[0]):
		V0[q] = init[q] 
	if p.On == 1:
		V0[-1] += 1.0 # End time initial guess
		
	# Convert constraints to lists
	lbg = np.zeros(vertcat(*g).shape[0]).tolist()
	ubg = np.zeros(vertcat(*g).shape[0]).tolist()
	lbx = vertcat(*lbx)
	ubx = vertcat(*ubx)

	nlp = {'x':vertcat(*V),'f':f, 'g':vertcat(*g)}
	
	options = {}
	options['warn_initial_bounds'] = True
	options['ipopt.linear_solver'] = 'ma27'
	options['ipopt.max_iter'] = 1000
	solver = nlpsol('solver','ipopt', nlp, options)

	sol = solver(x0=V0, lbx=lbx, ubx=ubx, lbg=lbg, ubg=ubg)
	
	# Extract all of the decision variables from the solved nlp
	v_opt = sol['x']
	if p.On == 1:
		tf = v_opt[-1]
		print tf*p.TF
	else:
		tf = p.TF

	[X_out,Z_out] = MonteCarlo(p,v_opt,nvar)

	return v_opt,X_out,Z_out,
	
[v_opt,X_out,Z_out] = StochOpt(p,v_opt,nvar,BO_P6,BO_Z3_g,BO_eta);

[back_off_eta, back_off_Z3_g,back_off_P_6,mu_eta,mu_Z3_g,mu_P_6,sigma_eta,sigma_Z3_g,sigma_P_6] = backoff(X_out,Z_out);


