from math import pi
from casadi import *
import numpy as np
import matplotlib.pyplot as plt
from collections import OrderedDict
from helpers import casadi_vec, casadi_struct, casadi_vec2struct, casadi_struct2vec
from Parameters import Parameters # Import the parameters
from Model import Model # Import the model equations
from Layout import Layout # Import the integrator, structure and timegrid 
from Initial import Initial # Import the initial guess

p = Parameters();
[f,dx,obj,x,z,u,t_end] = Model(p);
[Int,dae,dae,dae_x,dae_z,dae_u,tau_root,dae_t_end] = Layout(p,f,dx,obj,x,z,u,t_end);

def demopt(dae_x,dae_z,dae_u,dae,p,z):
	# Import collocation script
	from collocation_tools import simpleColl
	if p.On == 1:
		from collocation_tools_t import simpleColl
	else:
		from collocation_tools import simpleColl
	collfun = simpleColl(dae,tau_root,1.0/p.N)
	collfun = collfun.expand()
	# Number of varibles
	nx = dae["x"].shape[0] # Differential variables
	nu = dae["p"].shape[0] # Control variables
	nz = dae["z"].shape[0] # Algebraic variables
	nvar = nx + nx*p.d + nz*p.d + nu # number of variables 
	nv = nx*(p.N+1) + nx*p.d*p.N + nz*p.d*p.N + nu*p.N # Total number of variables
	
	Xs  = [MX.sym("X",nx) for i in range(p.N+1)]
	XCs = [MX.sym("XC",nx,p.d) for i in range(p.N)]
	Zs  = [MX.sym("Z",nz,p.d) for i in range(p.N)]
	Us  = [MX.sym("U",nu) for i in range(p.N)]
	if p.On == 1:
		t_end = MX.sym("t_end")
	V_block = OrderedDict()
	V_block["X"]  = Sparsity.dense(nx,1)
	V_block["XC"] = Sparsity.dense(nx,p.d)
	V_block["Z"]  = Sparsity.dense(nz,p.d)
	V_block["U"]  = Sparsity.dense(nu,1)
	
	# Simple bounds on states
	lbx = [];
	ubx = [];

	# Objective function
	f = 0.0;

	# List of constraints
	g = [];
	
	# List of all decision variables (determines ordering)
	V = [];
	
	for k in range(p.N):
		# Add decision variables
		V += [casadi_vec(V_block,X=Xs[k],XC=XCs[k],Z=Zs[k],U=Us[k])]
		
		if k==0:
			# Bounds at t=0
			x_lb = casadi_vec(dae_x,0.0,Pr=1.0,eta=1.0)
			x_ub = casadi_vec(dae_x,1.0,Pr=1.0,eta=1.0)
			xc_lb = casadi_vec(dae_x,0.0,eta=0.80)
			xc_lb = repmat(xc_lb,1,p.d)
			xc_ub = casadi_vec(dae_x,1.0)
			xc_ub = repmat(xc_ub,1,p.d)
			u_lb = casadi_vec(dae_u,0.0,P2=0.2,h=0.5)
			u_ub = casadi_vec(dae_u,1.0,h=2.5)
			z_lb = casadi_vec(dae_z,0.0,P6=0.48)
			z_lb = repmat(z_lb,1,p.d)
			z_ub = casadi_vec(dae_z,1.0,Z5_o=0.01,Z3_g=0.1)
			z_ub = repmat(z_ub,1,p.d)
			lbx.append(casadi_vec(V_block,-inf,X=x_lb,XC=xc_lb,Z=z_lb,U=u_lb))
			ubx.append(casadi_vec(V_block,inf,X=x_ub,XC=xc_ub,Z=z_ub,U=u_ub))
		else:
			# Bounds for other t
			x_lb = casadi_vec(dae_x,0.0,eta=0.80)
			x_ub = casadi_vec(dae_x,1.0)
			xc_lb = casadi_vec(dae_x,0.0,eta=0.80)
			xc_lb = repmat(xc_lb,1,p.d)
			xc_ub = casadi_vec(dae_x,1.0)
			xc_ub = repmat(xc_ub,1,p.d)
			u_lb = casadi_vec(dae_u,0.0,P2=0.2,h=0.5)
			u_ub = casadi_vec(dae_u,1.0,h=2.5)
			z_lb = casadi_vec(dae_z,0.0,P6=0.48)
			z_lb = repmat(z_lb,1,p.d)
			z_ub = casadi_vec(dae_z,1.0,Z5_o=0.01,Z3_g=0.1)
			z_ub = repmat(z_ub,1,p.d)
			lbx.append(casadi_vec(V_block,-inf,X=x_lb,XC=xc_lb,Z=z_lb,U=u_lb))
			ubx.append(casadi_vec(V_block,inf,X=x_ub,XC=xc_ub,Z=z_ub,U=u_ub))
		# Obtain collocation expressions
		if p.On == 1:
			# Varying time horizon
			[Xnext,constr,quad_k] = collfun(Xs[k],XCs[k],Zs[k],Us[k],t_end)
		else:
			# Fixed time horizon
			[Xnext,constr,quad_k] = collfun(Xs[k],XCs[k],Zs[k],Us[k])
		g.append(constr) # Add algebraic constraints
		g.append(Xs[k+1] - Xnext) # Add continuity constraint added
		# Add to the objective function
		if p.PV == 1:
			# Present Value is included
			f += quad_k/(1.05)**(k*t_end/12.0)
		else:
			# Present Value is not included
			f += quad_k
		
	V += [Xs[-1]] 
	
	if p.On == 1:
		# Add time horizon as varible for varying time horizin
		V += [t_end]
		f *= V[-1] 
		
	# Bounds for final t
	x_lb = casadi_vec(dae_x,0.0,eta=0.80)
	x_ub = casadi_vec(dae_x,1.0)
	lbx.append(x_lb)
	ubx.append(x_ub)
	
	if p.On == 1:
		# Add constraint on end time horizon
		t_lb = casadi_vec(dae_t_end,0.8)
		t_ub = casadi_vec(dae_t_end,1.2)
		lbx.append(t_lb)
		ubx.append(t_ub)
	
	#Initial guess
	init = Initial(p,z,V_block);
	if p.On == 1:
		V0 = np.zeros(init.shape[0]+1)
	else:
		V0 = np.zeros(init.shape[0])
		
	for q in range(init.shape[0]):
		V0[q] = init[q] 
	if p.On == 1:
		V0[-1] += 1.0 # End time initial guess
		
	# Convert constraints to lists
	lbg = np.zeros(vertcat(*g).shape[0]).tolist()
	ubg = np.zeros(vertcat(*g).shape[0]).tolist()
	lbx = vertcat(*lbx)
	ubx = vertcat(*ubx)

	nlp = {'x':vertcat(*V),'f':f, 'g':vertcat(*g)}
	
	options = {}
	options['warn_initial_bounds'] = True
	options['ipopt.linear_solver'] = 'ma27'
	options['ipopt.max_iter'] = 1000
	solver = nlpsol('solver','ipopt', nlp, options)

	sol = solver(x0=V0, lbx=lbx, ubx=ubx, lbg=lbg, ubg=ubg)
	
	# Extract all of the decision variables from the solved nlp
	v_opt = sol['x']
	if p.On == 1:
		tf = v_opt[-1]
		print tf*p.TF
	else:
		tf = p.TF
	return v_opt,nvar,lbx,ubx,tf
	
	
# [v_opt,nvar,lbx,ubx,tf] = demopt(dae_x,dae_z,dae_u,dae,p,z);
